from typing import List

import pytest
import torch
from torch import nn

import heavyball
import heavyball.utils
from benchmark.utils import get_optim
from heavyball.utils import set_torch, clean


class Param(nn.Module):
    def __init__(self, size):
        super().__init__()
        self.weight = nn.Parameter(torch.randn(size))

    def forward(self, inp):
        return self.weight.mean() * inp


@pytest.mark.parametrize("opt", ['ForeachPSGDKron', 'ForeachPaLMPAdam'])
@pytest.mark.parametrize("method", ['qr', 'newtonschulz2', 'svd', 'eigh'])
@pytest.mark.parametrize("size", [(16, 16, 16, 16), (4, 4, 4, 4), (512, 1, 128), (32128, 768)])
@pytest.mark.parametrize("merge,split", [(False, False), (True, False), (True, True)])
def test_merge(opt, method, size: List[int], merge, split, depth: int = 2, iterations: int = 5,
               outer_iterations: int = 3):
    if 'soap' not in opt.lower() and method != 'qr':
        raise pytest.skip('Only SOAP supports `method` argument')
    clean()
    set_torch()

    opt = getattr(heavyball, opt)
    heavyball.utils.zeroth_power_mode = method

    for _ in range(outer_iterations):
        clean()
        model = nn.Sequential(*[Param(size) for _ in range(depth)]).cuda()
        # We don't know if merging will use more or less memory, but we do know that it shouldn't crash. This test is to check if it crashes
        o = get_optim(opt, model.parameters(), lr=1e-3, merge_dims=merge, split=split, max_precond_dim=256,
                      max_size_triangular=256)

        for i in range(iterations):
            model(torch.randn((1, size[0]), device='cuda')).sum().backward()
            o.step()
            o.zero_grad()
            print(o.state_size())

        del model, o



import heavyball
import heavyball.utils
import pytest
import torch
from benchmark.utils import get_optim
from heavyball.utils import clean, set_torch, PSGDBase
from torch import nn


def get_memory():
    clean()
    torch.cuda.synchronize()
    clean()
    torch.cuda.synchronize()
    return torch.cuda.memory_allocated()


@pytest.mark.parametrize("opt", heavyball.__all__)
@pytest.mark.parametrize("size,depth", [(256, 128)])
def test_foreach(opt, size, depth: int, iterations: int = 5, outer_iterations: int = 3):
    set_torch()

    opt = getattr(heavyball, opt)

    peaks = []
    losses = []

    for foreach in [True, False]:
        torch.manual_seed(0x2131290)
        peaks.append([])
        losses.append([])

        for i in range(outer_iterations):
            clean()
            model = nn.Sequential(*[nn.Linear(size, size) for _ in range(depth)]).cuda()
            clean()

            torch.cuda.reset_peak_memory_stats()
            torch.cuda.reset_max_memory_allocated()
            torch.cuda.reset_max_memory_cached()
            torch.cuda.reset_accumulated_memory_stats()

            clean()
            o = get_optim(opt, model.parameters(), lr=1e-3, foreach=foreach)
            clean()

            for _ in range(iterations):
                loss = model(torch.randn((1, size), device='cuda')).sum()
                loss.backward()
                o.step()
                o.zero_grad()
                losses[-1].append(loss.detach())

            del model, o
            clean()

            peak = torch.cuda.memory_stats()['allocated_bytes.all.peak']

            if i > 0:
                peaks[-1].append(peak)

    for p0, p1 in zip(*peaks):
        assert p0 > p1
    for l0, l1 in zip(*losses):  # increase error tolerance for PSGD, as we have different RNGs -> expected differences
        assert torch.allclose(l0, l1, rtol=0.01 if isinstance(opt, PSGDBase) else 1e-5)



import heavyball
import heavyball.utils
import pytest
import torch
from benchmark.utils import get_optim
from heavyball.utils import clean, set_torch
from torch import nn


def get_memory():
    clean()
    torch.cuda.synchronize()
    clean()
    torch.cuda.synchronize()
    return torch.cuda.memory_allocated()


@pytest.mark.parametrize("opt", ['ForeachPSGDKron', 'ForeachPaLMPAdam', 'ForeachPurePSGD', 'ForeachDelayedPSGD'])
@pytest.mark.parametrize("method",
                         ['norm_clip_', 'mu_law_compress', 'a_law_compress', 'trust_region_clip_', 'identity'])
@pytest.mark.parametrize("size,depth", [(128, 1), (16, 4)])
def test_clip(opt, method, size, depth: int, iterations: int = 100, outer_iterations: int = 3):
    set_torch()

    opt = getattr(heavyball, opt)

    for i in range(outer_iterations):
        model = nn.Sequential(*[nn.Linear(size, size) for _ in range(depth)]).cuda()

        torch.cuda.reset_peak_memory_stats()
        torch.cuda.reset_max_memory_allocated()
        torch.cuda.reset_max_memory_cached()
        torch.cuda.reset_accumulated_memory_stats()

        model_allocated = get_memory()
        o = get_optim(opt, model.parameters(), lr=1e-3, clip_fn=getattr(heavyball.utils, method))
        losses = torch.zeros((iterations,), device='cuda')
        for itr in range(iterations):
            src = torch.randn((4, size), device='cuda')
            tgt = src
            loss = (model(src) - tgt).square().mean()
            loss.backward()
            o.step()

            opt_allocated = get_memory()
            o.zero_grad()
            losses[itr] = loss

        del model, o

        arange = torch.arange(iterations, device='cuda', dtype=torch.float32)
        lwma_bwd = (losses @ torch.flip(arange, [0])).item()
        lwma_fwd = (losses @ arange).item()
        assert lwma_bwd > lwma_fwd

        peak = torch.cuda.memory_stats()['allocated_bytes.all.peak']

        print(f'Peak: {peak / model_allocated:.2f}x | Opt: {opt_allocated / model_allocated:.2f}x')
        if i > 0:
            assert peak / model_allocated < v['peak']
            assert opt_allocated / model_allocated < v['after']
